#!/usr/bin/python

from contextlib import contextmanager
import os
import subprocess
import sys

import charmhelpers.core.hookenv as hookenv
from charmhelpers.core.host import service_restart, service_stop
from charmhelpers.fetch import apt_install, apt_update


hooks = hookenv.Hooks()


def is_postgres_available():
    """Is the relation with PostgreSQL ready for us to use?"""
    if not hookenv.is_relation_made("db"):
        hookenv.log("Missing relation to PostgreSQL")
        return False
    this_unit = hookenv.local_unit()
    allowed_units_str = hookenv.relation_get("allowed-units")
    if allowed_units_str is not None:
        allowed_units = allowed_units_str.split(" ")
    else:
        allowed_units = []
    if this_unit not in allowed_units:
        hookenv.log("Unit is not yet allowed (%s)" % " ".join(allowed_units))
        return False
    database_name = hookenv.relation_get("database")
    if database_name != "reviewboard":
        hookenv.log(
            "Database provided by PostgreSQL is not 'reviewboard': %s" % repr(
                database_name))
        return False
    return True


def is_rb_site_installed():
    """Return True if we have a site installed"""
    return os.path.exists("/var/www/reviewboard")


@contextmanager
def _django_environ():
    """Context manager for setting up DJANGO_SETTINGS_MODULE."""
    cwd = os.getcwd()
    config_dir = "/var/www/reviewboard/conf"
    os.chdir(config_dir)
    sys.path.insert(0, config_dir)
    settings_module = os.environ.get("DJANGO_SETTINGS_MODULE")
    os.environ["DJANGO_SETTINGS_MODULE"] = "reviewboard.settings"
    yield
    if settings_module is None:
        del os.environ["DJANGO_SETTINGS_MODULE"]
    else:
        os.environ["DJANGO_SETTINGS_MODULE"] = settings_module
    sys.path[:] = sys.path[1:]
    os.chdir(cwd)


def install_rb_site():
    """Gather and use details for configuring the ReviewBoard instance."""
    config = hookenv.config()
    if not is_postgres_available():
        return False
    password = config.get("admin-password")
    if password is None:
        hookenv.log("admin-password is not set, refusing to configure site")
        return False
    hookenv.log("PostgreSQL relation is ready!")
    public_address = hookenv.unit_get("public-address")
    host = config.get("host", "localhost")
    if host == "localhost":
        host = public_address
    hookenv.log("Installing rb-site for %s in /var/www/reviewboard" % host)
    db_relation_data = hookenv.relation_get()
    args = ["--noinput",
            "--domain-name=%s" % host,
            "--db-type=postgresql",
            "--db-name=reviewboard",
            "--db-host=%s:%s" % (
                db_relation_data["host"], db_relation_data["port"]),
            "--db-user=%s" % db_relation_data["user"],
            "--db-pass=%s" % db_relation_data["password"],
            "--cache-type=memcached",
            "--web-server-type=apache",
            "--web-server-port=80",
            "--python-loader=wsgi",
            "--admin-user=admin",
            "--admin-password=%s" % password,
            "/var/www/reviewboard"]
    subprocess.check_call(["rb-site", "install"] + args)
    hookenv.log("Installed rb-site!")
    return True


def configure_apache2():
    """Set up Apache site, removing default and adding our one."""
    hookenv.log("Configuring Apache...")
    subprocess.check_call([
        "cp", "/var/www/reviewboard/conf/apache-wsgi.conf",
        "/etc/apache2/sites-available/reviewboard.conf"])
    subprocess.check_call(["a2dissite", "000-default"])
    subprocess.check_call(["a2ensite", "reviewboard.conf"])
    for path in ["/var/www/reviewboard/htdocs/media/uploaded",
                 "/var/www/reviewboard/htdocs/media/ext",
                 "/var/www/reviewboard/htdocs/static/ext",
                 "/var/www/reviewboard/data"]:
        subprocess.check_call(["chown", "-R", "www-data", path])
    service_restart("apache2")
    hookenv.log("Apache configured!")


@hooks.hook('db-relation-joined')
def db_relation_joined():
    hookenv.relation_set(relation_settings={"database": "reviewboard"})


@hooks.hook('db-relation-changed')
def db_relation_changed():
    if not is_rb_site_installed():
        install_rb_site()
    if is_rb_site_installed():
        configure_apache2()


@hooks.hook("website-relation-joined")
def website_relation_joined():
    private_ip = hookenv.unit_get("private-address")
    hookenv.relation_set(
        relation_settings={"hostname": private_ip, "port": 80})


@hooks.hook("config-changed")
def config_changed():
    config = hookenv.config()
    if not is_rb_site_installed():
        install_rb_site()
    if is_rb_site_installed():
        configure_apache2()
    else:
        # Failed to install rb site for some reason, try again later.
        return
    if config.get("admin-password"):
        with _django_environ():
            from django.contrib.auth.management.commands import changepassword
            command = changepassword.Command()
            command._get_pass = lambda *args: config["admin-password"]
            command.execute("admin")
    if config.get("host"):
        if config["host"] == "localhost":
            host = hookenv.unit_get("public-address")
        else:
            host = config["host"]
        with _django_environ():
            from django.contrib.sites.models import Site
            cur_site = Site.objects.get_current()
            cur_site.domain = host
            cur_site.save()


@hooks.hook("install")
def install():
    apt_update(fatal=True)
    apt_install(["apache2-mpm-prefork",
                 "libapache2-mod-wsgi",
                 "python-setuptools",
                 "python-dev",
                 "memcached",
                 "python-memcache",
                 "patch",
                 "python-psycopg2"], fatal=True)
    hookenv.log("Installing ReviewBoard")
    subprocess.check_call(["easy_install", "ReviewBoard"])


@hooks.hook("start")
def start():
    service_restart("apache2")
    hookenv.open_port(80)


@hooks.hook("stop")
def stop():
    hookenv.close_port(80)
    service_stop("apache2")


def main(argv):
    try:
        hooks.execute(argv)
    except hookenv.UnregisteredHookError as e:
        hookenv.log('Unknown hook {} - skipping.'.format(e))


if __name__ == '__main__':
    main(sys.argv)
